"""
Repeats the last word of the last message in the conversation, and use it in
an annoying “C’est toi le” sentence.

Installation
------------

You only have to load the plugin:

.. code-block:: none

    /load stoi

.. glossary::

    /stoi
        **Usage:** ``/stoi``

"""
from plugin import BasePlugin
import tabs
import string
import xhtml

char_we_dont_want = string.punctuation+' ’„“”…«»'

class Plugin(BasePlugin):
    def init(self):
        for tab_type in (tabs.MucTab, tabs.PrivateTab, tabs.ConversationTab):
            self.api.add_tab_command(tab_type, 'stoi',
                                     handler=self.stoi,
                                     help="Repeats the last word of the last message "
                                          "in the conversation, and use it in an "
                                          "annoying “C’est toi le” sentence.",
                                     short='C’est toi le stoi.')

    def stoi(self, args):
        messages = self.api.get_conversation_messages()
        if not messages:
            # Do nothing if the conversation doesn’t contain any message
            return
        last_message = messages[-1]
        txt = xhtml.clean_text(last_message.txt)
        for char in char_we_dont_want:
            txt = txt.replace(char, ' ')
        if txt.strip():
            last_word = txt.split()[-1]
        else:
            last_word = "vide"
        self.api.send_message("C’est toi le %s." % last_word)



"""
This plugin lets you execute a system command through poezio.

Installation
------------

You only have to load the plugin. ::

    /load exec

Usage
-----

.. warning:: Running commands that start a daemon or an interface is not a good
             idea.

.. glossary::

    /exec
        **Usage:** ``/exec [-o|-O] <command>``

        Execute a system command.

        ::

            /exec command

        Will give you the result in the information buffer.

        ::

            /exec -o command

        Will send the result of the command into the current tab, if possible.

        ::

            /exec -O command

        Will send the result of the command and the command summary into the current
        tab, if possible.

"""

from plugin import BasePlugin
import os
import common
import shlex
import subprocess

class Plugin(BasePlugin):
    def init(self):
        self.api.add_command('exec', self.command_exec,
                usage='[-o|-O] <command>',
                help='Execute a shell command and prints the result in the information buffer. The command should be ONE argument, that means it should be between \"\". The first argument (before the command) can be -o or -O. If -o is specified, it sends the result in the current conversation. If -O is specified, it sends the command and its result in the current conversation.\nExample: /exec -O \"uptime\" will send “uptime\n20:36:19 up  3:47,  4 users,  load average: 0.09, 0.13, 0.09” in the current conversation.',
                short='Execute a command')

    def command_exec(self, args):
        args = common.shell_split(args)
        if len(args) == 1:
            command = args[0]
            arg = None
        elif len(args) == 2:
            command = args[1]
            arg = args[0]
        else:
            self.api.run_command('/help exec')
            return
        try:
            process = subprocess.Popen(['sh', '-c', command], stdout=subprocess.PIPE)
        except OSError as e:
            self.api.information('Failed to execute command: %s' % (e,), 'Error')
            return
        result = process.communicate()[0].decode('utf-8')
        if arg and arg == '-o':
            if not self.api.send_message('%s' % (result,)):
                self.api.information('Cannot send result (%s), this is not a conversation tab' % result)
        elif arg and arg == '-O':
            if not self.api.send_message('%s:\n%s' % (command, result)):
                self.api.information('Cannot send result (%s), this is not a conversation tab' % result)
        else:
            self.api.information('%s:\n%s' % (command, result), 'Info')
        return



import os
import logging
from sys import version_info

from sleekxmpp.plugins.xep_0048 import *
from common import safeJID
from config import config

log = logging.getLogger(__name__)

def iter(xml, tag=''):
    if version_info[1] >= 2:
        return xml.iter(tag)
    else:
        return xml.getiterator(tag)

preferred = config.get('use_bookmarks_method', 'pep').lower()
if preferred not in ('pep', 'privatexml'):
    preferred = 'privatexml'
not_preferred = 'privatexml' if preferred == 'pep' else 'privatexml'
methods = ('local', preferred, not_preferred)


class Bookmark(object):
    possible_methods = methods

    def __init__(self, jid, name=None, autojoin=False, nick=None, password=None, method='privatexml'):
        self.jid = jid
        self.name = name or jid
        self.autojoin = autojoin
        self.nick = nick
        self.password = password
        self._method = method

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        if value not in self.possible_methods:
            log.debug('Could not set bookmark storing method: %s', value)
            return
        self._method = value

    def __repr__(self):
        return '<%s%s%s>' % (self.jid, ('/'+self.nick) if self.nick else '', '|autojoin' if self.autojoin else '')

    def stanza(self):
        """
        Generate a <conference/> stanza from the instance
        """
        el = Conference()
        el['name'] = self.name
        el['jid'] = self.jid
        el['autojoin'] = 'true' if self.autojoin else 'false'
        if self.nick:
            el['nick'] = self.nick
        if self.password:
            el['password'] = self.password
        return el

    def local(self):
        """Generate a str for local storage"""
        local = self.jid
        if self.nick:
            local += '/%s' % self.nick
        local += ':'
        if self.password:
            config.set_and_save('password', self.password, section=self.jid)
        return local

    def parse_from_element(el, method=None):
        """
        Generate a Bookmark object from a <conference/> element
        """
        jid = el.get('jid')
        name = el.get('name')
        autojoin = True if el.get('autojoin', 'false').lower() in ('true', '1') else False
        nick = None
        for n in iter(el, 'nick'):
            nick = nick.text
        password = None
        for p in iter(el, 'password'):
            password = p.text

        return Bookmark(jid, name, autojoin, nick, password, method)

bookmarks = []

def get_by_jid(value):
    """
    Get a bookmark by bare jid
    """
    for item in bookmarks:
        if item.jid == value:
            return item

def remove(value):
    """
    Remove a bookmark (with its jid or directly the Bookmark object).
    """
    if isinstance(value, str):
        value = get_by_jid(value)
    bookmarks.remove(value)

def stanza_storage(method):
    """Generate a <storage/> stanza with the conference elements."""
    storage = Bookmarks()
    for b in filter(lambda b: b.method == method, bookmarks):
        storage.append(b.stanza())
    return storage

def save_pep(xmpp):
    """Save the remote bookmarks via PEP."""
    xmpp.plugin['xep_0048'].set_bookmarks(stanza_storage('pep'),
            method='xep_0223')

def save_privatexml(xmpp):
    """"Save the remote bookmarks with privatexml."""
    xmpp.plugin['xep_0048'].set_bookmarks(stanza_storage('privatexml'),
            method='xep_0049')

def save_remote(xmpp, method=preferred):
    """Save the remote bookmarks."""
    method = "privatexml" if method != 'pep'  else 'pep'

    try:
        if method is 'privatexml':
            xmpp.plugin['xep_0048'].set_bookmarks(stanza_storage('privatexml'),
                    method='xep_0049')
        else:
            xmpp.plugin['xep_0048'].set_bookmarks(stanza_storage('pep'),
                    method='xep_0223')
    except:
        import traceback
        log.debug("Could not save the bookmarks:\n%s" % traceback.format_exc())
        return False
    return True

def save_local():
    """Save the local bookmarks."""
    all = ''.join(bookmark.local() for bookmark in bookmarks if bookmark.method is 'local')
    config.set_and_save('rooms', all)

def save(xmpp, core=None):
    """Save all the bookmarks."""
    save_local()
    if config.get('use_remote_bookmarks', 'true').lower() != 'false':
        preferred = config.get('use_bookmarks_method', 'privatexml')
        if not save_remote(xmpp, method=preferred) and core:
            core.information('Could not save bookmarks.', 'Error')
            return False
        elif core:
            core.information('Bookmarks saved', 'Info')
    return True

def get_pep(xmpp):
    """Add the remotely stored bookmarks via pep to the list."""
    try:
        iq = xmpp.plugin['xep_0048'].get_bookmarks(method='xep_0223', block=True)
    except:
        return False
    for conf in iter(iq.xml, '{storage:bookmarks}conference'):
        b = Bookmark.parse_from_element(conf, method='pep')
        if not get_by_jid(b.jid):
            bookmarks.append(b)
    return True

def get_privatexml(xmpp):
    """Add the remotely stored bookmarks via privatexml to the list."""
    try:
        iq = xmpp.plugin['xep_0048'].get_bookmarks(method='xep_0049', block=True)
    except:
        return False
    for conf in iter(iq.xml, '{storage:bookmarks}conference'):
        b = Bookmark.parse_from_element(conf, method='privatexml')
        if not get_by_jid(b.jid):
            bookmarks.append(b)
    return True

def get_remote(xmpp):
    """Add the remotely stored bookmarks to the list."""
    if xmpp.anon:
        return
    method = config.get('use_bookmarks_method', '')
    if not method:
        pep, privatexml = True, True
        for method in methods[1:]:
            if method == 'pep':
                pep = get_pep(xmpp)
            else:
                privatexml = get_privatexml(xmpp)
        if pep and not privatexml:
            config.set_and_save('use_bookmarks_method', 'pep')
        elif privatexml and not pep:
            config.set_and_save('use_bookmarks_method', 'privatexml')
        elif not pep and not privatexml:
            config.set_and_save('use_bookmarks_method', '')
    else:
        if method == 'pep':
            get_pep(xmpp)
        else:
            get_privatexml(xmpp)

def get_local():
    """Add the locally stored bookmarks to the list."""
    rooms = config.get('rooms', '')
    if not rooms:
        return
    rooms = rooms.split(':')
    for room in rooms:
        jid = safeJID(room)
        if jid.bare == '':
            continue
        if jid.resource != '':
            nick = jid.resource
        else:
            nick = None
        passwd = config.get_by_tabname('password', '', jid.bare, fallback=False) or None
        b = Bookmark(jid.bare, autojoin=True, nick=nick, password=passwd, method='local')
        if not get_by_jid(b.jid):
            bookmarks.append(b)



"""
Plugin manager module.
Define the PluginManager class, the one that glues all the plugins and
the API together. Defines also a bunch of variables related to the
plugin env.
"""

import imp
import os
from os import path
import sys
import logging
from gettext import gettext as _
from sys import version_info

import core
import tabs
from plugin import PluginAPI
from config import config

log = logging.getLogger(__name__)

load_path = []

plugins_dir = config.get('plugins_dir', '')
plugins_dir = plugins_dir or\
    os.path.join(os.environ.get('XDG_DATA_HOME') or\
                     os.path.join(os.environ.get('HOME'), '.local', 'share'),
                 'poezio', 'plugins')
plugins_dir = os.path.expanduser(plugins_dir)

plugins_conf_dir = config.get('plugins_conf_dir', '')
if not plugins_conf_dir:
    config_home = os.environ.get('XDG_CONFIG_HOME')
    if not config_home:
        config_home = os.path.join(os.environ.get('HOME'), '.config')
    plugins_conf_dir = os.path.join(config_home, 'poezio', 'plugins')
plugins_conf_dir = os.path.expanduser(plugins_conf_dir)

try:
    os.makedirs(plugins_dir)
except OSError:
    pass
else:
    load_path.append(plugins_dir)

try:
    os.makedirs(plugins_conf_dir)
except OSError:
    pass

default_plugin_path = path.join(path.dirname(path.dirname(__file__)), 'plugins')

if os.path.exists(default_plugin_path):
    load_path.append(default_plugin_path)

try:
    import poezio_plugins
except:
    pass
else:
    if poezio_plugins.__path__:
        load_path.append(list(poezio_plugins.__path__)[0])

if version_info[1] >= 3: # 3.3 & >
    from importlib import machinery
    finder = machinery.PathFinder()

class PluginManager(object):
    """
    Plugin Manager
    Contains all the references to the plugins
    And keeps track of everything the plugin has done through the API.
    """
    def __init__(self, core):
        self.core = core
        self.modules = {} # module name -> module object
        self.plugins = {} # module name -> plugin object
        self.commands = {} # module name -> dict of commands loaded for the module
        self.event_handlers = {} # module name -> list of event_name/handler pairs loaded for the module
        self.tab_commands = {} #module name -> dict of tab types; tab type -> commands loaded by the module
        self.keys = {} # module name → dict of keys/handlers loaded for the module
        self.tab_keys = {} #module name → dict of tab types; tab type → list of keybinds (tuples)
        self.roster_elements = {}
        self.plugin_api = PluginAPI(core, self)

    def disable_plugins(self):
        for plugin in set(self.plugins.keys()):
            try:
                self.unload(plugin)
            except:
                pass

    def load(self, name, notify=True):
        """
        Load a plugin.
        """
        if name in self.plugins:
            self.unload(name)

        try:
            module = None
            if version_info[1] < 3: # < 3.3
                if name in self.modules:
                    imp.acquire_lock()
                    module = imp.reload(self.modules[name])
                else:
                    file, filename, info = imp.find_module(name, load_path)
                    imp.acquire_lock()
                    module = imp.load_module(name, file, filename, info)
            else: # 3.3 & >
                loader = finder.find_module(name, load_path)
                if not loader:
                    self.core.information('Could not find plugin')
                    return
                module = loader.load_module()

        except Exception as e:
            import traceback
            log.debug("Could not load plugin %s: \n%s", name, traceback.format_exc())
            self.core.information("Could not load plugin %s: %s" % (name, e), 'Error')
        finally:
            if version_info[1] < 3 and imp.lock_held():
                imp.release_lock()
            if not module:
                return

        self.modules[name] = module
        self.commands[name] = {}
        self.keys[name] = {}
        self.tab_keys[name] = {}
        self.tab_commands[name] = {}
        self.event_handlers[name] = []
        self.plugins[name] = module.Plugin(self.plugin_api, self.core, plugins_conf_dir)
        if notify:
            self.core.information('Plugin %s loaded' % name, 'Info')

    def unload(self, name, notify=True):
        if name in self.plugins:
            try:
                for command in self.commands[name].keys():
                    del self.core.commands[command]
                for key in self.keys[name].keys():
                    del self.core.key_func[key]
                for tab in list(self.tab_commands[name].keys()):
                    for command in self.tab_commands[name][tab][:]:
                        self.del_tab_command(name, getattr(tabs, tab), command[0])
                    del self.tab_commands[name][tab]
                for tab in list(self.tab_keys[name].keys()):
                    for key in self.tab_keys[name][tab][:]:
                        self.del_tab_key(name, getattr(tabs, tab), key[0])
                    del self.tab_keys[name][tab]
                for event_name, handler in self.event_handlers[name][:]:
                    self.del_event_handler(name, event_name, handler)

                self.plugins[name].unload()
                del self.plugins[name]
                del self.commands[name]
                del self.keys[name]
                del self.tab_commands[name]
                del self.event_handlers[name]
                if notify:
                    self.core.information('Plugin %s unloaded' % name, 'Info')
            except Exception as e:
                import traceback
                log.debug("Could not unload plugin: \n%s", traceback.format_exc())
                self.core.information("Could not unload plugin: %s" % e, 'Error')

    def add_command(self, module_name, name, handler, help, completion=None, short='', usage=''):
        """
        Add a global command.
        """
        if name in self.core.commands:
            raise Exception(_("Command '%s' already exists") % (name,))

        commands = self.commands[module_name]
        commands[name] = core.Command(handler, help, completion, short, usage)
        self.core.commands[name] = commands[name]

    def del_command(self, module_name, name):
        """
        Remove a global command added through add_command.
        """
        if name in self.commands[module_name]:
            del self.commands[module_name][name]
            if name in self.core.commands:
                del self.core.commands[name]

    def add_tab_command(self, module_name, tab_type, name, handler, help, completion=None, short='', usage=''):
        """
        Add a command only for a type of Tab.
        """
        commands = self.tab_commands[module_name]
        t = tab_type.__name__
        if name in tab_type.plugin_commands:
            return
        if not t in commands:
            commands[t] = []
        commands[t].append((name, handler, help, completion))
        tab_type.plugin_commands[name] = core.Command(handler, help, completion, short, usage)
        for tab in self.core.tabs:
            if isinstance(tab, tab_type):
                tab.update_commands()

    def del_tab_command(self, module_name, tab_type, name):
        """
        Remove a command added through add_tab_command.
        """
        commands = self.tab_commands[module_name]
        t = tab_type.__name__
        if not t in commands:
            return
        for command in commands[t]:
            if command[0] == name:
                commands[t].remove(command)
                del tab_type.plugin_commands[name]
                for tab in self.core.tabs:
                    if isinstance(tab, tab_type) and name in tab.commands:
                        del tab.commands[name]

    def add_tab_key(self, module_name, tab_type, key, handler):
        """
        Associate a key binding to a handler only for a type of Tab.
        """
        keys = self.tab_keys[module_name]
        t = tab_type.__name__
        if key in tab_type.plugin_keys:
            return
        if not t in keys:
            keys[t] = []
        keys[t].append((key, handler))
        tab_type.plugin_keys[key] = handler
        for tab in self.core.tabs:
            if isinstance(tab, tab_type):
                tab.update_keys()

    def del_tab_key(self, module_name, tab_type, key):
        """
        Remove a key binding added through add_tab_key.
        """
        keys = self.tab_keys[module_name]
        t = tab_type.__name__
        if not t in keys:
            return
        for _key in keys[t]:
            if _key[0] == key:
                keys[t].remove(_key)
                del tab_type.plugin_keys[key]
                for tab in self.core.tabs:
                    if isinstance(tab, tab_type) and key in tab.key_func:
                        del tab.key_func[key]

    def add_key(self, module_name, key, handler):
        """
        Associate a global key binding to a handler, except if it
        already exists.
        """
        if key in self.core.key_func:
            raise Exception(_("Key '%s' already exists") % (key,))
        keys = self.keys[module_name]
        keys[key] = handler
        self.core.key_func[key] = handler

    def del_key(self, module_name, key):
        """
        Remove a global key binding added by a plugin.
        """
        if key in self.keys[module_name]:
            del self.keys[module_name][key]
            if key in self.core.key_func:
                del self.core.commands[key]

    def add_event_handler(self, module_name, event_name, handler, position=0):
        """
        Add an event handler. If event_name isn’t in the event list, assume
        it is a sleekxmpp event.
        """
        eh = self.event_handlers[module_name]
        eh.append((event_name, handler))
        if event_name in self.core.events.events:
            self.core.events.add_event_handler(event_name, handler, position)
        else:
            self.core.xmpp.add_event_handler(event_name, handler)

    def del_event_handler(self, module_name, event_name, handler):
        """
        Remove an event handler if it exists.
        """
        if event_name in self.core.events.events:
            self.core.events.del_event_handler(None, handler)
        else:
            self.core.xmpp.del_event_handler(event_name, handler)
        eh = self.event_handlers[module_name]
        eh = list(filter(lambda e : e != (event_name, handler), eh))

    def completion_load(self, the_input):
        """
        completion function that completes the name of the plugins, from
        all .py files in plugins_dir
        """
        try:
            names = set()
            for path in load_path:
                try:
                    add = set(os.listdir(path))
                    names |= add
                except:
                    pass
        except OSError as e:
            self.core.information(_('Completion failed: %s' % e), 'Error')
            return
        plugins_files = [name[:-3] for name in names if name.endswith('.py')
                and name != '__init__.py' and not name.startswith('.')]
        plugins_files.sort()
        return the_input.new_completion(plugins_files, 1, '', quotify=False)

    def completion_unload(self, the_input):
        """
        completion function that completes the name of the plugins that are loaded
        """
        return the_input.new_completion(sorted(self.plugins.keys()), 1, '', quotify=False)

    def on_plugins_dir_change(self, new_value):
        global plugins_dir
        if plugins_dir in load_path:
            load_path.remove(plugins_dir)
        load_path.insert(0, new_value)
        plugins_dir = new_value

    def on_plugins_conf_dir_change(self, new_value):
        global plugins_conf_dir
        plugins_conf_dir = new_value



"""
Add a subtle little advertising in your messages.

Installation
------------

You only have to load the plugin:

.. code-block:: none

    /load spam


Configuration
-------------
[spam]
ad = I’m a happy poezio user. Get it at http://poezio.eu

"""

from plugin import BasePlugin

class Plugin(BasePlugin):
    def init(self):
        self.api.add_event_handler('muc_say', self.advert)
        self.api.add_event_handler('conversation_say', self.advert)
        self.api.add_event_handler('private_say', self.advert)

    def advert(self, msg, tab):
        msg['body'] = "%s\n\n%s" % (msg['body'], self.config.get("ad", "Sent from poezio"))



"""
Shuffle the words in every message you send in a :ref:`muctab`
(may confuse the reader).

Installation
------------

You only have to load the plugin:

.. code-block:: none

    /load shuffle

"""
from plugin import BasePlugin
from random import shuffle

class Plugin(BasePlugin):
    def init(self):
        self.api.add_event_handler('muc_say', self.shuffle)

    def shuffle(self, msg, tab):
        split = msg['body'].split()
        shuffle(split)
        msg['body'] = ' '.join(split)



"""
Repeats the last message in the conversation.

Installation
------------

You only have to load the plugin:

.. code-block:: none

    /load mirror

Command
-------

.. glossary::

    /mirror
        **Usage:** ``/mirror``

"""
from plugin import BasePlugin
import tabs

class Plugin(BasePlugin):
    def init(self):
        for tab_type in (tabs.MucTab, tabs.PrivateTab, tabs.ConversationTab):
            self.api.add_tab_command(tab_type, 'mirror',
                    handler=self.mirror,
                    help='Repeat the last message from the conversation.',
                    short='Repeat the last message from the conversation.')

    def mirror(self, args):
        messages = self.api.get_conversation_messages()
        if not messages:
            # Do nothing if the conversation doesn’t contain any message
            return
        last_message = messages[-1]
        self.api.send_message(last_message.txt)



"""
This plugin adds a message at 00:00 in each of your chat tabs saying that the
date has changed.

Installation
------------
You only have to load the plugin.::

    /load day_change

"""

from gettext import gettext as _
from plugin import BasePlugin
import datetime
import tabs
import timed_events

class Plugin(BasePlugin):
    def init(self):
        self.schedule_event()

    def cleanup(self):
        self.api.remove_timed_event(self.next_event)

    def schedule_event(self):
        day_change = datetime.datetime.combine(datetime.date.today(), datetime.time())
        day_change += datetime.timedelta(1)
        self.next_event = timed_events.TimedEvent(day_change, self.day_change)
        self.api.add_timed_event(self.next_event)

    def day_change(self):
        msg = _("Day changed to %s") % (datetime.date.today().isoformat())

        for tab in self.core.tabs:
            if (isinstance(tab, tabs.MucTab) or
                isinstance(tab, tabs.PrivateTab) or
                isinstance(tab, tabs.ConversationTab)):
                tab.add_message(msg)

        self.core.refresh_window()
        self.schedule_event()
